/*
 * Copyright (c) 1995 PFU Limited.
 *	author Osamu Satoh, Kazuhisa Shirakami
 */
package teikade.gui;

import java.awt.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Observer;
import java.util.Observable;
import teikade.sys.*;
import teikade.lang.*;
import teikade.util.Grep;
import teikade.util.TabConverter;
import teikade.gui.menu.TeikadeMenuBar;
import teikade.gui.component.*;
import teikade.gui.dialog.TeikadeDialog;
import teikade.gui.editor.*;
import teikade.debug.DebuggerManager;

public class TeikadeFrame extends DFrame implements Observer, EditDataObserver {

    private static Vector teikadeFrames = new Vector();

    protected CodePath	startup;
    protected CodePath	current;

    protected TeikadeMenuBar	menuBar;
    protected CodePathField	codePathField;
    private Button	  	acceptButton;
    private Button	  	resetButton;
    private Button	  	syncButton;
    protected CodeData	  	codeData;

    private boolean	noUpdateCode = false;

    // strings for BorderLayout
    protected final static String NORTH  = "North";
    protected final static String WEST   = "West";
    protected final static String SOUTH  = "South";
    protected final static String EAST   = "East";
    protected final static String CENTER = "Center";

    public TeikadeFrame(CodePath cp, String title) {
	super(title);
	setBackground(Color.lightGray);

	startup = cp;
	current = new CodePath(cp);

	menuBar = createMenuBar();
	setMenuBar(menuBar);
	add(CENTER, createDisplayItems());
	pack();

	teikadeFrames.addElement(this);
	TeikadeProperties.observers.addObserver(this);
	Transcript.getTranscriptMenuBar().addFrame(this);
    }

    protected TeikadeMenuBar createMenuBar() {
	return new TeikadeMenuBar();
    }

    protected Component createDisplayItems() {
	codePathField = new CodePathField();
	codePathField.showCodePath(current.printString());

	DTextArea codeView = new DTextArea(23, 81);
					    // Why? I'd like to set 24,80

	EditData editData = codeView.getEditData();
	editData.setObserver(this);
	editData.setStatusViewer(codePathField);
	editData.setEditable(false);

	menuBar.setEditData(editData);
	codeData = new CodeData(editData);

	Panel buttons = new Panel();
	buttons.setLayout(new FlowLayout(FlowLayout.CENTER, 2, 5));
	buttons.add(acceptButton = new DButton("Accept"));
	buttons.add(  syncButton = new DButton("Sync"  ));
	buttons.add( resetButton = new DButton("Reset" ));

	Panel operation = new Panel();
	operation.setLayout(new BorderLayout());
	operation.add(CENTER,  codePathField);
	operation.add(EAST, buttons);

	Panel code = new Panel();
	code.setLayout(new BorderLayout());
	code.add(CENTER, codeView);

	Panel panel = new Panel();
	panel.setLayout(new BorderLayout());
	panel.add("North", operation);
	panel.add("Center", code);
	return panel;
    }

    public void update(Observable observable, Object arg) {
	if (observable instanceof TeikadeProperties) {
	    updateProperties(this);
	}
	if (menuBar instanceof UseProperties) {
	    ((UseProperties)menuBar).updateProperties();
	}
	int num = menuBar.countMenus();
	for (int i = 0; i < num; i++) {
	    Menu menu = menuBar.getMenu(i);
	    if (menu instanceof UseProperties) {
		((UseProperties)menu).updateProperties();
	    }
	}
	invalidate();
	layout();
    }

    public void updateProperties(Container container) {
	Component comps[] = container.getComponents();
	for (int i = 0; i < comps.length; i++) {
	    Component comp = comps[i];
	    if (comp instanceof Container) {
		updateProperties((Container)comp);
	    } else if (comp instanceof UseProperties) {
		((UseProperties)comp).updateProperties();
	    }
	}
    }

    /**
     * update all of this frame
     */
    public static final void updateNotify(CodePath cp) {
	Enumeration en = teikadeFrames.elements();
	while (en.hasMoreElements()) {
	    TeikadeFrame frame = (TeikadeFrame)en.nextElement();
	    if (cp.hasInfluenceOn(frame.current)) {
		if (cp.matches(frame.current)
		||  cp.classManager() == null
		||  frame.current.classManager() == null) {
		    frame.update();
		} else {
		    frame.updateExceptCode();
		}
	    }
	}
    }

    protected final void updateExceptCode() {
	noUpdateCode = true;
	update();
	noUpdateCode = false;
    }

    /**
     * update all of this frame
     */
    protected void update() {
	updateCurrent();
    }

    /**
     * update CurrentMonitor, Menu and CodeData
     */
    protected final void updateCurrent() {
	codePathField.showCodePath(current.printString());
	if (!noUpdateCode) {
	    updateCode();
	}
	updateButtonActivity();
	updateMenuActivity();
    }

    /**
     * update CodeData, CurrentMonitor and Menus activity
     */
    protected void updateCode() {
	Manager manager = current.getManager();
	String code = manager.getCode();
	setCode(code, manager.isEditable());
    }

    protected void updateMenuActivity() {
	menuBar.updateActivity(this, current);
    }

    protected void updateButtonActivity() {
	updateButtonActivity(codeData.isEditable(), codeData.isEditable());
    }

    protected final void updateButtonActivity(boolean acceptable, boolean resetable) {
	if (acceptable) {
	    acceptButton.enable();
	} else {
	    acceptButton.disable();
	}
	if (resetable) {
	    resetButton.enable();
	} else {
	    resetButton.disable();
	}
	if (acceptable && current.getManager().isChanged()) {
	    syncButton.enable();
	} else {
	    syncButton.disable();
	}
    }

    /**
     * get code from CodeData
     */
    public final String getCode() {
	String code = codeData.getCode();
	if (current != null) {
	    ProjectManager projectManager = current.projectManager();
	    if (projectManager != null) {
		String projectName = projectManager.name();
		int fileTab = EditProperties.getFileTabWidth(projectName);
		int displayTab = EditProperties.hardTab;
		if (fileTab != displayTab) {
		    code = TabConverter.convert(code, displayTab, fileTab);
		}
	    }
	}
	return code;
    }

    /**
     * get selected code from CodeData
     */
    public final String getSelectedCode() {
	return codeData.getSelectedCode();
    }

    /**
     * set the code into CodeData
     */
    public final void setCode(String code) {
	setCode(code, false);
    }

    /**
     * set the code into CodeData
     */
    protected final void setCode(String code, boolean editable) {
	boolean codeSameAsView = code.equals(getCode());
	if (codeSameAsView) {
	    codeData.replaceOriginal(codeData.getCode());
	    codeData.setEditable(editable);
	    return;
	}
	if (current != null) {
	    ProjectManager projectManager = current.projectManager();
	    if (projectManager != null) {
		String projectName = projectManager.name();
		int fileTab = EditProperties.getFileTabWidth(projectName);
		int displayTab = EditProperties.hardTab;
		if (fileTab != displayTab) {
		    code = TabConverter.convert(code, fileTab, displayTab);
		}
	    }
	}
	if (codeData.isChanged()) {
	    if (!reset("This `" + current.printShortString() +
			      " has been changed, too.\nForce reset it?")) {
		codeData.replaceOriginal(code);
		return;
	    }
	}
	codeData.setEditable(editable);
	codeData.setCode(code);
	return;
    }

    /**
     * action handler
     */
    public boolean action(Event ev, Object arg) {
	try {
	    if (ev.target instanceof MenuItem) {
		menuBar.action(ev, arg, this, current);
		return true;
	    } else if (ev.target == acceptButton) {
		accept();
		return true;
	    } else if (ev.target == resetButton) {
		reset();
		return true;
	    } else if (ev.target == syncButton) {
		sync();
		return true;
	    }
	}
	catch (TeikadeException e) {
	    doReset();
	    current.classManager((ClassManager)null);
	    TeikadeFrame.updateNotify(current);
	    return true;
	}
	return super.action(ev, arg);
    }

    /**
     * accept code
     */
    public boolean acceptWithConfirm() {
	if (!codeData.isChanged()) {
	    return true;
	} else {
	    int chosen = (new TeikadeDialog(this,
				"'" + current.printShortString() +
				"' has been changed.\n" +
				"Accept it ? Or reset it?",
				null, "Accept", "Reset", "Cancel")).choose();
	    if (chosen == 3) {	// Cancel
		return false;
	    } else if (chosen == 2) {
		doReset();
		return true;
	    } else {
		return doAccept();
	    }
	}
    }

    /**
     * accept code
     */
    public boolean accept() {
	if (!codeData.isChanged()) return true;
	return doAccept();
    }

    /**
     * accept code
     *     return same CodePath, if success to accept as replaced
     *     return different CodePath, if success to accept as created new class/method
     *     return null, if fail to accept
     */
    protected final boolean doAccept() {

	String code = getCode();

	Manager oldManager = current.getManager();
	Manager newManager;
	try {
	    newManager = oldManager.setCode(code);
	}
	catch (TeikadeException e) {
	    information(e.getMessage());
	    return false;
	}
	if (newManager == null) return false;	// fail to accept

	// success to accept
	codeData.replaceOriginal(code);

	if (newManager != oldManager) {
	    current.setManager(newManager);
	}

	TeikadeFrame.updateNotify(current);
	return true;
    }

    /**
     * reset code
     */
    public boolean reset() {
	return reset("`" + current.printShortString() +
		     "' has been changed. Force reset it?");
    }

    /**
     * reset code
     */
    protected final boolean reset(String confirmMessage) {
	if (!codeData.isChanged()) {
	    return true;
	} else if (confirmMessage != null &&
		   !confirm(confirmMessage, "Reset")) {
	    return false;
	} else {
	    doReset();
	    return true;
	}
    }

    /**
     * reset code
     */
    protected final void doReset() {
	codeData.resetCode();
    }

    /**
     * sync code
     */
    public void sync() {
	sync(current);
    }

    /**
     * sync code
     */
    public void sync(CodePath cp) {
	String saveLabel = syncButton.getLabel();
	codePathField.viewStatus("Syncing... ", false);
	cp.getManager().sync();
	codePathField.viewStatus("Syncing... done.", false);
	TeikadeFrame.updateNotify(cp);
    }

    public void rebuild(CodePath cp) {
	String names[] = current.names();
	if (cp.packageManager() != null) {
	    cp.packageManager().rebuild();
	} else if (cp.projectManager() != null) {
	    cp.projectManager().rebuild();
	}
	current = new CodePath(names);	// for keep previous selection
	TeikadeFrame.updateNotify(cp);
    }

    public boolean updateEditData(EditData editData, int command) {
	switch (command) {
	case EditDataObserver.ACCEPT : return accept();
	case EditDataObserver.SYNC   : sync(current); return true;
	case EditDataObserver.RESET  : return reset();
	case EditDataObserver.COMPILE: return compile();
	case EditDataObserver.EXECUTE: return execute();
	case EditDataObserver.CLOSE  : return close();
	case EditDataObserver.QUIT   : return quit();
	case EditDataObserver.SPAWN  : spawn(startup); return true;
	case EditDataObserver.SPAWNNARROW: spawn(current); return true;
	case EditDataObserver.SYNCALL: sync(current.projectCodePath()); return true;
	}
	return false;
    }

    public boolean compile() {
	if (current.methodManager() != null) {
	    return compile(current.classCodePath(), false);
	} else {
	    return compile(current, false);;
	}
   }

    /**
     * compile
     */
    public boolean compile(CodePath cp, boolean updated) {
	if (codeData.isChanged()) {
	    int chosen = (new TeikadeDialog(this,
				"'" + current.printShortString() +
				"' has been changed.\n" +
				"Accept it ?",
				null,
				"Accept and compile",
				"No accept and compile",
				"Cancel")).choose();
	    if (chosen == 3) {	// Cancel
		return false;
	    } else if (chosen == 1) {
		if (doAccept() == false) {
		    return false;
		}
	    }
	}
	cp.getManager().sync();
	CompileFrame.openOn(cp, updated);
	return true;
    }

    public boolean execute() {
	ClassManager classManager = current.classManager();
	if (classManager == null) return false;
	return classManager.execute(null, currentDebugger());
    }

    /**
     * grep
     */
    public final void grep(CodePath cp) {
	String pattern = codeData.getSelectedCode();
	if (pattern == null || pattern.length() == 0) {
	    pattern = request(Grep.getInstance().getPatternLabel());
	}
	try {
	    grep(cp, pattern, "Occurences of: '" + pattern + "'" );
	}
	catch (TeikadeException e) {
	    information("grep: " + e.getMessage());
	}
    }

    public static final void grep(CodePath cp, String pattern, String title)
    throws TeikadeException {
	if (pattern == null || pattern.length() == 0) {
	    return;
	}
	GrepFrame.openOn(cp, pattern, title);
    }

    /**
     * spawn browser
     */
    public void spawn(CodePath cp) {
	Browser.openOn(cp, current);
    }

    /**
     * spawn hierarchy browser
     */
    public final void spawnHierarchy(CodePath cp) {
	CodePath listCp = new CodePath(cp);
	if (cp.methodManager() != null) {
	    listCp = cp.classCodePath();
	} else {
	    listCp = cp.packageCodePath();
	}
	HierarchyBrowser.openOn(listCp, cp,
			      "Hierarchy Browser on: " + cp.printShortString());
    }

    /**
     * spawn implementers
     */
    public final void spawnImplementers(CodePath cp) {
	spawnImplementers(cp, "Implementers of: " + cp.printString());
    }

    /**
     * spawn implementers
     */
    public final void spawnImplementers(CodePath cp, String title) {
	String regexp;
	try {
	    if (cp.methodManager() == null) {
		regexp = "implements[ \t]*([a-zA-Z_][a-zA-Z_0-9]*,[ \t]*)*" + cp.className();
		UsersFrame.openOn( cp.projectCodePath(), regexp, title );
	    } else {
		ImplementersFrame.openOn(cp.methodManager());
	    }
	}
	catch (TeikadeException e) {
	    information("grep: " + e.getMessage());
	}
    }

    /**
     * spawn implementers
     */
    public final void spawnImplementers(CodePath cp, IdentifierSpec idspec, String label) {
	try {
	    if (cp.methodManager() == null) {
		// not implemented yet
		String regexp = "implements[ \t]*([a-zA-Z_][a-zA-Z_0-9]*,[ \t]*)*" + cp.className();
		UsersFrame.openOn( cp.projectCodePath(), regexp, "Implementers of: " + label );
	    } else {
		ImplementersFrame.openOn(cp, idspec, label);
	    }
	}
	catch (TeikadeException e) {
	    information("grep: " + e.getMessage());
	}
    }

    /**
     * spawn senders
     */
    public final void spawnSenders(CodePath cp) {
	try {
	    if (cp.methodManager() == null) {
		String regexp = "[^a-zA-Z_0-9]" + cp.className() + "[^a-zA-Z_0-9]";
		grep(cp.projectCodePath(),  regexp, "Senders of: " + cp.className());
	    } else {
		SendersFrame.openOn(cp);
	    }
	}
	catch (TeikadeException e) {
	    information("grep: " + e.getMessage());
	}
    }

    /**
     * spawn messages
     */
    public final void spawnMessages(CodePath cp, String methodName) {
	try {
	    grep(cp, methodName, "Implementers of: " + methodName);
	}
	catch (TeikadeException e) {
	    information("grep: " + e.getMessage());
	}
    }

    /**
     * spawn users
     */
    public final void spawnUsers(CodePath cp) {
	try {
	    String title = "Users of: " + cp.printString();
	    if (cp.methodManager() == null) {
		String regexp = "[^a-zA-Z_0-9]" + cp.className() + "[^a-zA-Z_0-9]";
		UsersFrame.openOn(cp.projectCodePath(), regexp, title );
	    } else {
		spawnImplementers(cp, title);
	    }
	}
	catch (TeikadeException e) {
	    information("grep: " + e.getMessage());
	}
    }

    protected Object clone() {
	try {
	    return super.clone();
	} catch (CloneNotSupportedException e) {
	    return null;
	}
    }

    public void open() {
	show();
	update();
    }

    public boolean close() {
	if (!acceptWithConfirm()) return false;
	menuBar.close();
	int index = teikadeFrames.indexOf(this);
	teikadeFrames.removeElementAt(index);
	Transcript.getTranscriptMenuBar().removeFrameAt(index);
	TeikadeProperties.observers.deleteObserver(this);
	super.close();
	return true;
    }

    public boolean quit() {
	Transcript.getTranscript().close();
	return true;
    }

    final boolean tryClose() {
	close();
	return !isVisible();
    }

    final static Enumeration allFrames() {
	return teikadeFrames.elements();
    }

    /**
     * information message to user
     */
    public final void information(String message) {
	show();
	TeikadeDialog.information(this, message);
    }

    /**
     * request user input for String
     */
    public final String request(String message) {
	show();
	return TeikadeDialog.request(this, message, "");
    }

    /**
     * request user input for String
     */
    public final String request(String message, String def) {
	show();
	return TeikadeDialog.request(this, message, def);
    }

    /**
     * confirm Ok or Cancel
     */
    public final boolean confirm(String message) {
	show();
	return TeikadeDialog.confirm(this, message);
    }

    /**
     * confirm <label> or Cancel
     */
    public final boolean confirm(String message, String label) {
	show();
	return TeikadeDialog.confirm(this, message, label);
    }

    public DebuggerManager currentDebugger() {
        return null;
    }
}
