/*
 * Copyright (c) 1996 PFU Limited.
 *	author Kazuhisa Shirakami
 */
package teikade.gui.component;

import java.awt.*;
import java.util.Hashtable;
import java.util.Enumeration;
import teikade.sys.TeikadeProperties;

public class SplitLayout implements LayoutManager {
    public static final boolean HORIZONTAL = false;
    public static final boolean VERTICAL   = true;
    public static final boolean FOLLOW_ONE = false;
    public static final boolean FOLLOW_ALL = true;
    
    private boolean direction;		// HORIZONTAL or VERTICAL
    private boolean following;		// FOLLOW_ONE or FOLLOW_ALL
    private final Hashtable sizes = new Hashtable(10); // width or height
    private final Hashtable names = new Hashtable(10);

    public SplitLayout(boolean direction) {
	this(direction, FOLLOW_ONE);
    }

    public SplitLayout(boolean direction, boolean following) {
	this.direction = direction;
	this.following = following;
    }

    public boolean getDirection() {
	return direction;
    }

    public void addLayoutComponent(String name, Component comp) {
	if (name != null) {
	    names.put(comp, name);
	    int size = TeikadeProperties.getInteger(name);
	    if (size > 0) setSize(comp, size);
	}
    }

    public void removeLayoutComponent(Component comp) {
	sizes.remove(comp);
    }

    public Dimension preferredLayoutSize(Container parent) {
	Dimension dim = new Dimension(0, 0);
	int num = parent.countComponents();
	if (direction == HORIZONTAL) {
	    for (int i = 0; i < num; i++) {
		Component c = parent.getComponent(i);
		if (c.isVisible()) {
		    Dimension d = c.preferredSize();
		    dim.height = Math.max(dim.height, d.height);
		    dim.width += d.width;
		}
	    }
	} else {
	    for (int i = 0; i < num; i++) {
		Component c = parent.getComponent(i);
		if (c.isVisible()) {
		    Dimension d = c.preferredSize();
		    dim.width = Math.max(dim.width, d.width);
		    dim.height += d.height;
		}
	    }
	}
	Insets insets = parent.insets();
	dim.width += insets.left + insets.right;
	dim.height += insets.top + insets.bottom;
	return dim;
    }

    public Dimension minimumLayoutSize(Container parent) {
	Dimension dim = new Dimension(0, 0);
	int num = parent.countComponents();

	if (direction == HORIZONTAL) {
	    for (int i = 0; i < num; i++) {
		Component c = parent.getComponent(i);
		if (c.isVisible()) {
		    Dimension d = c.minimumSize();
		    dim.height = Math.max(dim.height, d.height);
		    dim.width += d.width;
		}
	    }
	} else {
	    for (int i = 0; i < num; i++) {
		Component c = parent.getComponent(i);
		if (c.isVisible()) {
		    Dimension d = c.minimumSize();
		    dim.width = Math.max(dim.width, d.width);
		    dim.height += d.height;
		}
	    }
	}
	Insets insets = parent.insets();
	dim.width += insets.left + insets.right;
	dim.height += insets.top + insets.bottom;
	return dim;
    }

    public void layoutContainer(Container target)  {
	layoutContainer(target, null);
    }

    public synchronized void layoutContainer(Container target, Component ignore)  {
	int num = target.countComponents();
	Dimension dim = target.size();
	Insets insets = target.insets();
	int width  = dim.width  - (insets.left + insets.right );
	int height = dim.height - (insets.top  + insets.bottom);
	Component last = null;
	int x = insets.left;
	int y = insets.top;
	if (direction == HORIZONTAL) {
	    for (int i = 0; i < num; i++) {
		Component c = target.getComponent(i);
		if (c.isVisible()) {
		    int w = getSize(c);
		    if (c != ignore) {
			c.reshape(x, y, w, height);
			if (c instanceof Container) ((Container)c).validate();
		    }
		    x += w;
		    last = c;
		}
	    }
	    if (last != null) {
		int w = dim.width - last.location().x - insets.right;
		if (w > 0) last.resize(w, height);
		if (last instanceof Container) ((Container)last).validate();
	    }
	} else {
	    for (int i = 0; i < num; i++) {
		Component c = target.getComponent(i);
		if (c.isVisible()) {
		    int h = getSize(c);
		    if (c != ignore) {
			c.reshape(x, y, width, h);
			if (c instanceof Container) ((Container)c).validate();
		    }
		    y += h;
		    last = c;
		}
	    }
	    if (last != null) {
		int h = dim.height - last.location().y - insets.bottom;
		if (h > 0) last.resize(width, h);
		if (last instanceof Container) ((Container)last).validate();
	    }
	}
    }

    private void setSize(Component comp, int size) {
	sizes.put(comp, new Integer(size));
    }

    private int getSize(Component comp) {
	Object size = sizes.get(comp);
	if (size != null) {
	    return ((Integer)size).intValue();
	} else {
	    Dimension dim = comp.preferredSize();
	    return direction == HORIZONTAL ? dim.width : dim.height;
	}
    }

    public synchronized void handle(Container parent, SplitHandle target,
				    int x, int y, boolean changeFollowing) {
	int num = parent.countComponents() - 1;
	for (int i = 1; i < num; i++) {
	    Component comp = parent.getComponent(i);
	    if (comp == target) {
		int delta, currentSize, followSize;
		boolean followOne = following ^ changeFollowing == FOLLOW_ONE;
		Component current = parent.getComponent(i - 1);
		Component follow = parent.getComponent(i + 1);
		if (direction == HORIZONTAL) {
		    delta = x;
		    currentSize = current.size().width;
		    followSize = followOne ? follow.size().width
				    	   : (parent.size().width - follow.location().x - parent.insets().right);
		} else {
		    delta = y;
		    currentSize = current.size().height;
		    followSize = followOne ? follow.size().height
					   : (parent.size().height - follow.location().y - parent.insets().bottom);
		}
		if (delta <= -currentSize) {
		    delta = -currentSize + 1;
		} else if (followSize <= delta) {
		    delta = followSize - 1;
		}
		if (followOne) setSize(follow, followSize - delta);
		setSize(current, currentSize + delta);
		layoutContainer(parent, target);
		return;
	    }
	}
    }

    public synchronized void handleDone(Container parent, SplitHandle target,
				    int x, int y, boolean changeFollowing) {
	//handle(parent, target, x, y, changeFollowing);
	Enumeration en = names.keys();
	while (en.hasMoreElements()) {
	    Object comp = en.nextElement();
	    Object size = sizes.get(comp);
	    if (size != null) {
		TeikadeProperties.putWithoutNotify(names.get(comp).toString(),
					      "" + ((Integer)size).intValue());
	    }
	}
    }
}
