/*
 * Copyright (c) 1999 PFU Limited.
 *	author Osamu Satoh
 */
package teikade.sys;

/**
 * The interface for user plugin commands of teikade.
 * <P>
 * Plugin commands are registered by adding a plugin entry
 * to the teikade properties file (.teikade/.properties).
 * <P>
 * The format of entries is: <BR>
 * <BLOCKQUOTE>
 * <CODE>
 * plugin.<I>name</I>=<I>PluginClassName</I>
 * </CODE>
 * </BLOCKQUOTE>
 * <P>
 * Plugin commands are instantiated only once for each command,
 * when the teikade system is started, and used for each invocation.
 * This means the plugin command instance itself must not
 * hold any runtime information to be initialized on every invocation.
 * For such information, the plugin command should instantiate
 * other objects which holds them when invoke method is called.
 */
public interface PluginCommand {
    /**
     * Initializes the command.
     * <P>
     * This method will be called only once just after PluginManager
     * instantiates the plugin command, before it registers the command.
     *
     * @param name Name of command
     * @exception PluginException something occurered during initialization
     */
    public void initialize(String name) throws PluginException;

    /**
     * Invokes the command.
     * <P>
     * @param name Name of command to which this command bounded
     * @param invoker the object which invokes this command.
     *  Usually, a TeikadeFrame in which command was invoked.
     * @param target the CodePath which was selected when this command was
     *  invoked. It's null, if none was selected.
     */
    public void invoke(String name, Object invoker, CodePath target)
	throws PluginException;

    /**
     * Checks whether this command can be invoked <B>now</B> for given target,
     * name and invoker.
     * @see invoke
     */
    public boolean isInvocableFor(String name, Object invoker, CodePath target);
}
